import setuptools


setuptools.setup(
    name="py2sqlm-andlvovsky",
    version="1.0.1",
    author="Viktor Ilvovskyi",
    author_email="viktorilvovskyi@gmail.com",
    description="Python 3 to PostgreSQL mapper",
    long_description="Python 3 to PostgreSQL mapper",
    long_description_content_type="text/markdown",
    url="https://github.com/AndLvovSky/metaprogramming/tree/lab3",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)
